@@ -35,20 +35,22 @@ Optimization-based Operations:
 """  
 
 
-def affine_map( T, P, t=None ):
+def affine_map( T, P, t=None , get_inverse=True):
     """
     Returns the affine map of a polytope.
     """        
     if type(t)==type(None):
-            t=np.zeros((P.n,1))
+            t=np.zeros((T.shape[0],1))
     if P.type=='AH_polytope':
         return pp.AH_polytope(t=t+np.dot(T,P.t),T=np.dot(T,P.T),P=P.P)
     elif P.type=='zonotope':
         return pp.zonotope(x=t+np.dot(T,P.x),G=np.dot(T,P.G))
-    elif P.type=="H_polytope":
-        if T.shape[0]>=T.shape[1]:
+    elif P.type=="H_polytope": 
+        if T.shape[0]>=T.shape[1] and get_inverse:
             Tinv=np.linalg.pinv(T)
             H=np.dot(P.H,Tinv)
+#            print("inverse error=",np.linalg.norm(np.dot(Tinv,T)-np.eye(T.shape[1])))
+            assert np.linalg.norm(np.dot(Tinv,T)-np.eye(T.shape[1]))<=1e-2*P.n
             return pp.H_polytope(H=H,h=P.h+np.dot(H,t))
         else:
             Q=pp.to_AH_polytope(P)
@@ -135,84 +137,134 @@ def check_non_empty(Q,tol=10**-5,solver="gurobi"):
         result=MP.Solve(prog)
     return result.is_success()
 
-def directed_Hausdorff_distance(Q1,Q2,ball="infinty_norm",solver="gurobi"):
-    r"""
-    Computes the directed Hausdorff distance of Q_1 and Q_2 (AH_polytopes)
-    ***************************************************************************
-    The optimization problem is:
-                        Minimize    epsilon  
-                        such that   Q1 \subset Q2+epsilon(Ball)
-                        
-    It is zero if and only if Q1 subset Q2. The method is based on 
-                
-                    Sadraddini&Tedrake, 2019, CDC (available on ArXiv)
-                    
-    We solve the following problem:
-        D*ball+Q1 subset Q2
-    We solve the following linear program:
-    ..math::
-        \min     D
-        s.t.    Lambda_1 H_1=H_2 Gamma_1
-                Lambda_2 H_1=H_ball Gamma_2
-                Lambda_1 h_1<=h_2 + H_2 beta_1
-                Lambda_2 h_2<=D h_ball + H_ball beta_2
-                x_2 - X_2 beta_1 - beta_2 = x_1
-                X_2 Gamma_1 + Gamma_2 = X_1
-    ***************************************************************************
-    """
-    Q1,Q2=pp.to_AH_polytope(Q1),pp.to_AH_polytope(Q2)
-    n=Q1.t.shape[0]
-    if ball=="infinty_norm":
-        HB=np.vstack((np.eye(n),-np.eye(n)))
-        hB=np.vstack((np.ones((n,1)),np.ones((n,1))))
-    elif ball=="l1":
-        HB,hb=make_ball(ball)
+#def directed_Hausdorff_distance(Q1,Q2,ball="infinty_norm",solver="gurobi"):
+#    r"""
+#    Computes the directed Hausdorff distance of Q_1 and Q_2 (AH_polytopes)
+#    ***************************************************************************
+#    The optimization problem is:
+#                        Minimize    epsilon  
+#                        such that   Q1 \subset Q2+epsilon(Ball)
+#                        
+#    It is zero if and only if Q1 subset Q2. The method is based on 
+#                
+#                    Sadraddini&Tedrake, 2019, CDC (available on ArXiv)
+#                    
+#    We solve the following problem:
+#        D*ball+Q1 subset Q2
+#    We solve the following linear program:
+#    ..math::
+#        \min     D
+#        s.t.    Lambda_1 H_1=H_2 Gamma_1
+#                Lambda_2 H_1=H_ball Gamma_2
+#                Lambda_1 h_1<=h_2 + H_2 beta_1
+#                Lambda_2 h_2<=D h_ball + H_ball beta_2
+#                x_2 - X_2 beta_1 - beta_2 = x_1
+#                X_2 Gamma_1 + Gamma_2 = X_1
+#    ***************************************************************************
+#    """
+#    Q1,Q2=pp.to_AH_polytope(Q1),pp.to_AH_polytope(Q2)
+#    n=Q1.t.shape[0]
+#    if ball=="infinty_norm":
+#        HB=np.vstack((np.eye(n),-np.eye(n)))
+#        hB=np.vstack((np.ones((n,1)),np.ones((n,1))))
+#    elif ball=="l1":
+#        HB,hb=make_ball(ball)
+#    prog=MP.MathematicalProgram()
+#    # Variables
+#    D=prog.NewContinuousVariables(1,1,"D")
+#    Lambda_1=prog.NewContinuousVariables(Q2.P.H.shape[0],Q1.P.H.shape[0],"Lambda_1")
+#    Lambda_2=prog.NewContinuousVariables(HB.shape[0],Q1.P.H.shape[0],"Lambda2")
+#    Gamma_1=prog.NewContinuousVariables(Q2.P.H.shape[1],Q1.P.H.shape[1],"Gamma1")
+#    Gamma_2=prog.NewContinuousVariables(HB.shape[1],Q1.P.H.shape[1],"Gamma1")
+#    beta_1=prog.NewContinuousVariables(Q2.P.H.shape[1],1,"beta1")
+#    beta_2=prog.NewContinuousVariables(HB.shape[1],1,"beta1")
+#    # Constraints
+#    # Lambda_1 and Lambda_2 positive
+#    prog.AddBoundingBoxConstraint(0,np.inf,Lambda_1)
+#    prog.AddBoundingBoxConstraint(0,np.inf,Lambda_2)
+#    # Lambda_1 H_1
+#    Lambda_H_Gamma(prog,Lambda_1,Q1.P.H,Q2.P.H,Gamma_1)
+#    # Lambda_2 H_1
+#    Lambda_H_Gamma(prog,Lambda_2,Q1.P.H,HB,Gamma_2)
+#    # Lambda_1 h_1
+#    Lambda_h_Inequality(prog,Lambda_1,beta_1,Q2.P.H,Q1.P.h,Q2.P.h)
+#    # Lambda_2 h_1
+#    Lambda_h_Inequality_D(prog,Lambda_2,beta_2,HB,Q1.P.h,hB,D)
+#    # X2 beta_1   
+#    prog.AddLinearEqualityConstraint(-np.hstack((Q2.T,np.eye(n))),Q1.t-Q2.t,np.vstack((beta_1,beta_2)))
+#    # X2 Gamma_1
+#    Aeq=np.hstack((Q2.T,np.eye(Q2.T.shape[0])))
+#    for i in range(Gamma_1.shape[1]):
+#        beq=Q1.T[:,i]
+#        var=np.hstack((Gamma_1[:,i],Gamma_2[:,i]))
+#        prog.AddLinearEqualityConstraint(Aeq,beq,var)
+#    # Cost
+#    # Optimize
+#    if solver=="gurobi":
+#            prog.AddLinearCost(D[0,0])
+#            result=gurobi_solver.Solve(prog,None,None)
+#    elif solver=="osqp":
+#        prog.AddQuadraticCost(D[0,0]*D[0,0])
+#        result=OSQP_solver.Solve(prog,None,None)
+#    else:
+#        result=MP.Solve(prog)
+#    if result.is_success():
+#        return np.asscalar(result.GetSolution(D))
+#    else:
+#        print("Optimization Failed")
+#
+#def old_Hausdorff_distance(Q1,Q2,directed=False,ball="infinty_norm",solver="gurobi"):
+#    return max(directed_Hausdorff_distance(Q1,Q2,ball,solver),directed_Hausdorff_distance(Q2,Q1,ball,solver))
+    
+def Hausdorff_distance(Q1,Q2,directed=False,ball="infinty_norm",solver="gurobi",k=-1):
+    X,Y=pp.to_AH_polytope(Q1),pp.to_AH_polytope(Q2)
     prog=MP.MathematicalProgram()
     # Variables
-    D=prog.NewContinuousVariables(1,1,"D")
-    Lambda_1=prog.NewContinuousVariables(Q2.P.H.shape[0],Q1.P.H.shape[0],"Lambda_1")
-    Lambda_2=prog.NewContinuousVariables(HB.shape[0],Q1.P.H.shape[0],"Lambda2")
-    Gamma_1=prog.NewContinuousVariables(Q2.P.H.shape[1],Q1.P.H.shape[1],"Gamma1")
-    Gamma_2=prog.NewContinuousVariables(HB.shape[1],Q1.P.H.shape[1],"Gamma1")
-    beta_1=prog.NewContinuousVariables(Q2.P.H.shape[1],1,"beta1")
-    beta_2=prog.NewContinuousVariables(HB.shape[1],1,"beta1")
-    # Constraints
-    # Lambda_1 and Lambda_2 positive
-    prog.AddBoundingBoxConstraint(0,np.inf,Lambda_1)
-    prog.AddBoundingBoxConstraint(0,np.inf,Lambda_2)
-    # Lambda_1 H_1
-    Lambda_H_Gamma(prog,Lambda_1,Q1.P.H,Q2.P.H,Gamma_1)
-    # Lambda_2 H_1
-    Lambda_H_Gamma(prog,Lambda_2,Q1.P.H,HB,Gamma_2)
-    # Lambda_1 h_1
-    Lambda_h_Inequality(prog,Lambda_1,beta_1,Q2.P.H,Q1.P.h,Q2.P.h)
-    # Lambda_2 h_1
-    Lambda_h_Inequality_D(prog,Lambda_2,beta_2,HB,Q1.P.h,hB,D)
-    # X2 beta_1   
-    prog.AddLinearEqualityConstraint(-np.hstack((Q2.T,np.eye(n))),Q1.t-Q2.t,np.vstack((beta_1,beta_2)))
-    # X2 Gamma_1
-    Aeq=np.hstack((Q2.T,np.eye(Q2.T.shape[0])))
-    for i in range(Gamma_1.shape[1]):
-        beq=Q1.T[:,i]
-        var=np.hstack((Gamma_1[:,i],Gamma_2[:,i]))
-        prog.AddLinearEqualityConstraint(Aeq,beq,var)
-    # Cost
-    # Optimize
-    if solver=="gurobi":
-            prog.AddLinearCost(D[0,0])
-            result=gurobi_solver.Solve(prog,None,None)
-    elif solver=="osqp":
-        prog.AddQuadraticCost(D[0,0]*D[0,0])
-        result=OSQP_solver.Solve(prog,None,None)
+    n=Q1.n
+    D1=prog.NewContinuousVariables(1,"D1")
+    D2=prog.NewContinuousVariables(1,"D2")
+    if ball=="infinty_norm":
+        P_ball=pp.unitbox(n).H_polytope
+    elif ball in ["L1",1,"1","l1"]:
+        P_ball=pp.unitball(n,1)
     else:
-        result=MP.Solve(prog)
+        print("I don't recognize the ball norm")
+        raise NotImplementedError    
+        
+    if P_ball.type=='H_polytope':
+        Dball1=pp.H_polytope(P_ball.H,P_ball.h*D1)
+        if not directed:
+            Dball2=pp.H_polytope(P_ball.H,P_ball.h*D2)
+            
+    if P_ball.type=='AH_polytope':
+        Dball1=pp.AH_polytope(t=P_ball.t*D1,T=P_ball.T,\
+                             P=pp.H_polytope(P_ball.P.H,P_ball.P.h*D1))
+        if not directed: 
+            Dball2=pp.AH_polytope(t=P_ball.t*D2,T=P_ball.T,\
+                     P=pp.H_polytope(P_ball.P.H,P_ball.P.h*D2))
+
+    X_plus=pp.minkowski_sum(X,Dball1)
+    pp.subset(prog,Y,X_plus,k=k)    
+    prog.AddLinearCost(np.array([1]),np.array([0]),D1)
+    if not directed:
+        Y_plus=pp.minkowski_sum(Y,Dball2)
+        pp.subset(prog,X,Y_plus,k=k)
+        prog.AddLinearCost(np.array([1]),np.array([0]),D2)
+    if solver=="gurobi":
+        result=gurobi_solver.Solve(prog,None,None)    
     if result.is_success():
-        return np.asscalar(result.GetSolution(D))
+        dXY=np.asscalar(result.GetSolution(D1))
+        if not directed:
+            dYX=np.asscalar(result.GetSolution(D2))
+            return max(dXY,dYX),dXY,dYX
+        else:
+            return dXY
 
-def Hausdorff_distance(Q1,Q2,ball="infinty_norm",solver="gurobi"):
-    return max(directed_Hausdorff_distance(Q1,Q2,ball,solver),directed_Hausdorff_distance(Q2,Q1,ball,solver))
-    
 def distance_polytopes(Q1,Q2,ball="infinity",solver="gurobi"):
+    """
+    Finds the closest two points in two polytopes and their distance.
+    It is zero if polytopes have non-empty intersection
+    """
     Q1,Q2=pp.to_AH_polytope(Q1),pp.to_AH_polytope(Q2)
     n=Q1.n
     prog=MP.MathematicalProgram()
@@ -358,6 +410,7 @@ def bounding_box(Q,solver="Gurobi"):
     a=np.zeros((Q.n,1))
     # Lower Corners
     for i in range(Q.n):
+#        print(i,"lower")
         e=c.evaluator()
         a[i,0]=1
         e.UpdateCoefficients(a.reshape(Q.n))
@@ -367,6 +420,7 @@ def bounding_box(Q,solver="Gurobi"):
         a[i,0]=0
     # Upper Corners
     for i in range(Q.n):
+#        print(i,"upper")
         e=c.evaluator()
         a[i,0]=-1
         e.UpdateCoefficients(a.reshape(Q.n))
@@ -554,10 +608,22 @@ def intersection(P1,P2):
         h=np.vstack((X.P.h,Y.P.h-np.linalg.multi_dot([Y.P.H,Ty_inv,X.t-Y.t])))
         new_P=pp.H_polytope(H,h)
         return pp.AH_polytope(T=T,t=X.t,P=new_P)
+    
+
 
     
 """
+*************************************************************
+*************************************************************
+*************************************************************
+*************************************************************
+
 Pydrake Mathematical Program Helper: Matrix based Constraints
+
+*************************************************************
+*************************************************************
+*************************************************************
+
 """
 def AddMatrixInequalityConstraint_classical(mathematical_program,A,X,B):
     raise NotImplementedError    
@@ -638,7 +704,7 @@ def boxing_order_reduction(zonotope,desired_order=1):
     inputs: input zonotope , order of the output zonotope
     output: zonotope
 
-    Based on Kopetzki, Anna-Kathrin, Bastian Schürmann, and Matthias Althoff.
+    Based on Kopetzki, Anna-Kathrin, Bastian Schurmann, and Matthias Althoff.
     "Methods for order reduction of zonotopes."
     2017 IEEE 56th Annual Conference on Decision and Control (CDC). IEEE, 2017.
     """
@@ -659,12 +725,12 @@ def boxing_order_reduction(zonotope,desired_order=1):
     
     elif dimension == desired_numberofcolumns:
         G_box = np.diag(np.sum(abs( G ) ,axis=1 ))
-        return pp.zonotope( G_box , x)  
+        return pp.zonotope( G=G_box , x=x)  
     
     else:
         G_reduced , G_untouched = sorting_generator( G , desired_numberofcolumns )
         G_box = np.concatenate(  ( np.diag(np.sum(abs( G_reduced ) ,axis=1 )) , G_untouched   ), axis=1  )
-        return pp.zonotope( G_box , x)  
+        return pp.zonotope( G=G_box , x=x)  
 
 
 def pca_order_reduction(zonotope,desired_order=1):
